/*=======================================================================*
 * A useless (?) program to convert a QL screen, not yet QXL compatible, *
 * from address $20000 (131072) to a PC PCX file                         *
 *-----------------------------------------------------------------------*
 * Copyright Norman Dunbar June 1994.                                    *
 *-----------------------------------------------------------------------*
 * BUGS OUTSTANDING.                                                     *
 *                                                                       *
 *-----------------------------------------------------------------------*
 * HISTORY.                                                              *
 *                                                                       *
 * 27/06/94 - First attempt.                                       v1.00 *
 * 25/08/94 - Add commandline parameters & wildcards.              v1.01 *
 *=======================================================================*/


#include <stdio_h>
#include <qdos_h>
#include <string_h>
#include <stdlib_h>

long (*_cmdchannels)() = NULL;                 /* No channels redirected */
long (*_stackchannels)() = NULL;               /* Or on stack either     */
int  (*_conwrite)() = NULL;                    /* Or output translation  */
void (*_cmdwildcard)() = cmdexpand;            /* Allow wildcard params  */

void consetup_title();                         /* Fancy title window */
void (*_consetup)() = consetup_title;          /* Link it in         */

char _prog_name[] = "Scr2pcx";                 /* Program name       */
char _version[] = "Version 1.01";              /* Version number     */
char _copyright[] = "Norman Dunbar, 1994";     /* Author & copright  */

unsigned char palette[48] = {  0, 0, 0,       /*  0 = Black      */
                               191, 0, 0,     /*  1 = Dk Red     */
                               0, 191, 0,     /*  2 = Dk Green   */
                               191, 191, 0,   /*  3 = Dk Yellow  */
                               0, 0, 191,     /*  4 = Dk Blue    */
                               191, 0, 191,   /*  5 = Dk Magenta */
                               0, 191, 191,   /*  6 = Dk Cyan    */
                               192, 192, 192, /*  7 = Lt Grey    */
                               128, 128, 128, /*  8 = Dk Grey    */
                               255, 0, 0,     /*  9 = Red        */
                               0, 255, 0,     /* 10 = Green      */
                               255, 255, 0,   /* 11 = Yellow     */
                               0, 0, 255,     /* 12 = Blue       */
                               255, 0, 255,   /* 13 = Magenta    */
                               0, 255, 255,   /* 14 = Cyan       */
                               255, 255, 255  /* 15 = White      */
                            };

void write_pcx_header(FILE *fp);               /* Write pcx header info  */
void load_screen(FILE *fp);                    /* Load QL screen file    */
void convert_screen(FILE *fp, int mode);       /* Convert it to _PCX     */
void set_mode(int mode);                       /* Check & set QL mode    */
void compress_pcx(unsigned char *buffer, FILE *fp); /* Compress a line   */


/*=====================================================================MAIN==*/
main(int argc, char *argv[])
{
   /*-------------------------------------------*
    * The program starts here for non C readers *
    *-------------------------------------------*/

   char pcx_file[128];     /* Buffer for QL screen filename   */
   char scr_file[128];     /* Buffer for PC BMP filename      */
   int  mode;              /* Holds 4 or 8 accordingly        */
   int  file;              /* Loop around all SCR files       */
   FILE *scr = NULL;       /* File pointer for QL screen file */
   FILE *pcx = NULL;       /* File pointer for PC PCX file    */


   /*----------------------------------------*
    * Did I get at least 3 (+1) parameters ? *
    *----------------------------------------*/

   if (argc < 4) {
       printf("USAGE : ex SCR2PCX;'scr_files mode pcx_device'\n\n");
       printf("Where :\n\n");
       printf("scr_files  = All filenames to be converted to bmp files\n");
       printf("             EG. flp1_*_SCR = all '_scr' files on 'flp1_'\n\n");
       printf("mode       = Screen mode of ALL the supplied filenames\n");
       printf("             EG. 4 or 8 for mode 4 or mode 8\n\n");
       printf("pcx_device = Device where all converted files are written\n");
       printf("             EG. ram1_ = write converted files to 'ram1_'\n\n");
       exit(0);
   }

   /*--------------------------*
    * Set the mode if required *
    *--------------------------*/

   mode = atoi(argv[argc -2]);
   if ((mode != 4) && (mode != 8)) {
       printf("Incorrect mode supplied, must be 4 or 8 only\n");
       exit (0);
   }

   set_mode(mode);

   /*---------------------------------------*
    * A loop to process all the input files *
    *---------------------------------------*/

   for (file = 1; file < argc -2; file++) {
       strcpy(scr_file, argv[file]);

       scr = fopen(scr_file, "r");
       if (scr == NULL) {
           sd_clear(fgetchid(stdout), -1);
           printf("%s can't be opened.\n", scr_file);
           continue;
       }

       strcpy(pcx_file, argv[argc-1]);
       strncat(pcx_file, scr_file +5, strlen(scr_file) -5);
       strcat(pcx_file, "_pcx");

       pcx = fopen(pcx_file, "wb");
       if (pcx == NULL) {
           sd_clear(fgetchid(stdout), -1);
           printf("%s can't be opened.\n", pcx_file);
           fclose(scr);
           continue;
       }

       /*-------------------------------------------------------*
        * Write out the various PC header stuff for a PCX file. *
        *-------------------------------------------------------*/

       write_pcx_header(pcx);

       /*-------------------------------------*
        * Convert the screen to a PCX format. *
        *-------------------------------------*/

       load_screen(scr);
       convert_screen(pcx, mode);

       fclose(pcx);
       fclose(scr);

   } /* end for */

   sd_clear(fgetchid(stdout), -1);
   exit(0);
}



/*=================================================================SET_MODE==*/
void set_mode(int mode)
{
   /*------------------------------------------------*
    * Read the current mode for the QL, if it is     *
    * different from that required for the screen    *
    * that is to be converted, set the correct mode. *
    *------------------------------------------------*/

   short s_mode,   /* Current mode */
         d_type;   /* Display type */


   /*------------------------------*
    * Read the current mode first. *
    *------------------------------*/

   d_type = -1;
   s_mode = -1;

   mt_dmode(&s_mode, &d_type);

   /*-------------------------------------------*
    * Is the current mode the same as we need ? *
    * if not, set it.                           *
    *-------------------------------------------*/

   if (!s_mode) s_mode = 4;    /* Zero = mode 4 */

   if (s_mode != (short)mode) {
       s_mode = mode;
       d_type = -1;
       mt_dmode(&s_mode, &d_type);
   }
}


/*==============================================================LOAD_SCREEN==*/
void load_screen(FILE *fp)
{
   /*-----------------------------------------*
    * Simply read 256 scan lines of 128 bytes *
    * into the QL display memory at 131072.   *
    *-----------------------------------------*/

   fread((char *)131072, 256, 128, fp);
}


/*=========================================================WRITE_PCX_HEADER==*/
void write_pcx_header(FILE *fp)
{
   static struct header {char  manufacturer;
                         char  version;
                         char  encoding;
                         char  bits_pixel;
                         short x_min;
                         short y_min;
                         short x_max;
                         short y_max;
                         short h_res;
                         short v_res;
                         char  palette[48];
                         char  reserved;
                         char  planes;
                         short bytes_line;
                         short palette_type;
                         char  filler[58];
                        } pcxh;

   memset((char *)&pcxh, 0, sizeof(struct header));
   pcxh.manufacturer = 0x0a;
   pcxh.version = 2;
   pcxh.encoding = 1;
   pcxh.bits_pixel = 4;
   pcxh.x_max = 0xff01;
   pcxh.y_max = 0xff00;
   pcxh.h_res = 0x0002;
   pcxh.v_res = 0x0001;

   for (pcxh.planes = 0; pcxh.planes < 48; pcxh.planes++)
       pcxh.palette[pcxh.planes] = palette[pcxh.planes];

   pcxh.planes =
   pcxh.bytes_line = 0x0001;
   pcxh.palette_type = 2;

   fwrite((char *)&pcxh, 1, sizeof(struct header), fp);

}


/*===========================================================CONVERT_SCREEN==*/
void convert_screen(FILE *fp, int mode)
{
   int scan_line,          /* Loop counter for each scan line      */
       line_offset,        /* Offset into display memory for line  */
       word,               /* Loop counter for each word in a line */
       word_offset,        /* Offset into memory for each word     */
       here;               /* Offset into uncompressed pixel data  */

   unsigned char   red,    /* Odd addresses byte of word offset    */
                   green,  /* Even addressed byte of word offset   */

                   top4_nibble[] = {0x00,  /* Black */
                                    0x90,  /* Red   */
                                    0xA0,  /* Green */
                                    0xF0}, /* White */

                   bot4_nibble[] = {0x00,  /* Black */
                                    0x09,  /* Red   */
                                    0x0A,  /* Green */
                                    0x0F}, /* White */

                   top8_nibble[] = {0x00,  /* Black   */
                                    0xC0,  /* Blue    */
                                    0x90,  /* Red     */
                                    0xD0,  /* Magenta */
                                    0xA0,  /* Green   */
                                    0xE0,  /* Cyan    */
                                    0xB0,  /* Yellow  */
                                    0xF0}, /* White   */

                   bot8_nibble[] = {0x00,  /* Black   */
                                    0x0C,  /* Blue    */
                                    0x09,  /* Red     */
                                    0x0D,  /* Magenta */
                                    0x0A,  /* Green   */
                                    0x0E,  /* Cyan    */
                                    0x0B,  /* Yellow  */
                                    0x0F}, /* White   */

                   pixel[8],    /* One word of pixel data  */
                   buffer[256]; /* One line of colour data */


   /*------------------------------------------*
    * There are 256 scan_lines in a QL screen  *
    * so calculate its position in memory. A   *
    * PCX file scans from top left down.       *
    *------------------------------------------*/

   for (scan_line = 0; scan_line < 256; scan_line++) {
       line_offset = 131072 + (scan_line * 128);
       memset(buffer, 0, 256);

       /*------------------------------------------*
        * There are 64 words in a scan_line. The   *
        * memory address is the scan_line address  *
        * plus the word * 2.                       *
        *------------------------------------------*/

       for (word = 0; word < 64; word++) {
           here = word << 2;
           word_offset = line_offset + (word << 1);

           /*------------------------------------------------------*
            * Fetch the green byte and the red byte for each word. *
            *------------------------------------------------------*/

           green = *((unsigned char *)word_offset);
           red   = *((unsigned char *)(word_offset +1));

           /*------------------------------------------*
            * invert the display to show where we are. *
            *------------------------------------------*/

           *((unsigned char *)word_offset) = mode == 4? ~green: (~green)& 0xAA;
           *((unsigned char *)(word_offset +1)) = ~red;

           /*========== MODE 4 ==========*=========== MODE 8 ===========*
            * GREEN | RED | PIXEL COLOUR * GR | RD | BL | PIXEL COLOUR  *
            *============================*==============================*
            *   0   |  0  | 00 = BLACK   *  0 |  0 |  0 | 000 = BLACK   *
            *   0   |  1  | 01 = RED     *  0 |  0 |  1 | 001 = BLUE    *
            *   1   |  0  | 10 = GREEN   *  0 |  1 |  0 | 010 = RED     *
            *   1   |  1  | 11 = WHITE   *  0 |  1 |  1 | 011 = MAGENTA *
            *----------------------------*  1 |  0 |  0 | 100 = GREEN   *
                                         *  1 |  0 |  1 | 101 = CYAN    *
                                         *  1 |  1 |  0 | 110 = YELLOW  *
                                         *  1 |  1 |  1 | 111 = WHITE   *
                                         *------------------------------*/

           /*--------------------------------------------------*
            * Using the above table, mask each bit out of the  *
            * red and green bytes, shift them along to the far *
            * right but leave the green bit in bit 1 while the *
            * red bit is in bit 0, now OR them together to get *
            * a value between 0 and 3 as above.                *
            *--------------------------------------------------*/

           if (mode == 4) {

               pixel[0] = ((green & 128) >> 6) | ((red & 128) >> 7);
               pixel[1] = ((green &  64) >> 5) | ((red &  64) >> 6);
               pixel[2] = ((green &  32) >> 4) | ((red &  32) >> 5);
               pixel[3] = ((green &  16) >> 3) | ((red &  16) >> 4);
               pixel[4] = ((green &   8) >> 2) | ((red &   8) >> 3);
               pixel[5] = ((green &   4) >> 1) | ((red &   4) >> 2);
               pixel[6] = ((green &   2))      | ((red &   2) >> 1);
               pixel[7] = ((green &   1) << 1) |  (red &   1);

               /*----------------------------------------------------*
                * In a PC PCX file, 16 colour, each pixel uses four  *
                * bits. The 4 bits make up the index into the colour *
                * pallette above. The array PIXEL holds data for 8   *
                * pixels so we can write out 4 bytes to the buffer.  *
                *----------------------------------------------------*/

               buffer[here + 0] = top4_nibble[pixel[0]] | bot4_nibble[pixel[1]];
               buffer[here + 1] = top4_nibble[pixel[2]] | bot4_nibble[pixel[3]];
               buffer[here + 2] = top4_nibble[pixel[4]] | bot4_nibble[pixel[5]];
               buffer[here + 3] = top4_nibble[pixel[6]] | bot4_nibble[pixel[7]];

           } else {

               pixel[0] = pixel[1] = ((green & 128) >> 5) | ((red & 192) >> 6);
               pixel[2] = pixel[3] = ((green &  32) >> 3) | ((red &  48) >> 4);
               pixel[4] = pixel[5] = ((green &   8) >> 1) | ((red &  12) >> 2);
               pixel[6] = pixel[7] = ((green &   2) << 1) | ((red &   3));

               buffer[here + 0] = top8_nibble[pixel[0]] | bot8_nibble[pixel[1]];
               buffer[here + 1] = top8_nibble[pixel[2]] | bot8_nibble[pixel[3]];
               buffer[here + 2] = top8_nibble[pixel[4]] | bot8_nibble[pixel[5]];
               buffer[here + 3] = top8_nibble[pixel[6]] | bot8_nibble[pixel[7]];

           } /* end if */
       } /* end for word  */

       compress_pcx(&buffer[0], fp);

   } /* end for scan_line */

}


/*=============================================================COMPRESS_PCX==*/
void compress_pcx(unsigned char *buffer, FILE *fp)
{
   /*--------------------------------------*
    * Compress a single scan line worth of *
    * bytes to the output PCX file.        *
    *--------------------------------------*/

   unsigned int same,          /* Count of 'same' bytes */
                this = 0;      /* Byte being examined   */

   do {

       /*---------------------------------------*
        * First count up to 63 identical bytes, *
        * but stay in the buffer.               *
        *---------------------------------------*/

       same = 0;
       while ((buffer[this + same] == buffer[this + same + 1]) &&
               ((this + same) < 256) && (same < 63))
               same++;

       /*----------------------------------------*
        * If we have a group of identical bytes, *
        * write a marker byte with the count in  *
        * the lower 5 bits, then the byte value. *
        * Set the pointer to the next byte to be *
        * checked.                               *
        *----------------------------------------*/

       if (same) {
           fputc(same | 0xc0, fp);     /* Count byte         */
           fputc(buffer[this], fp);    /* Byte value         */
           this += same;               /* Next byte to check */

       } else {

           /*----------------------------------------*
            * This must be a single value. If it is  *
            * greater than or equal to $C0 then it   *
            * must be marked as a sequence of 1 byte *
            * whoops, data compression ?             *
            *----------------------------------------*/

           if ((buffer[this] & 0xc0) == 0xc0) {
               fputc(0xc1, fp);        /* Count value = 1 */
           }

           /*------------------------------------*
            * Now simply write the byte value &  *
            * bump the pointer to the next byte. *
            *------------------------------------*/

           fputc(buffer[this], fp);   /* Byte value         */
           this++;                    /* Next byte to check */
       }
   } while (this < 256);
}


